
/*
 * $Id: opts.cc,v 1.1.1.9 1999/02/13 00:07:06 rousskov Exp $
 *
 *
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "polygraph.h"

#include "Assert.h"
#include "LogComment.h"
#include "UniqId.h"
#include "opts.h"
#include "gadgets.h"
#include "globals.h"


/* opts tools */

static const char *valStr(const char *val) { return val ? val : "<none>"; }

static
bool splitVal(const String &val, String &pref, String &suf, char del = ':') {
	if (val) {
		if (const char *h = strchr(val, del)) {
			pref = val(0, h-val);
			suf = h+1;
			return true;
		} else
			pref = val;
	}
	return false;
}

bool isTime(const char *str, Time &t) {
	// note: things are set up so that "sec" stands for "1sec" 
	//       (convenient for rates)
	const char *ptr = str;
	int i = 1;
	double d;

	// try to preserve integer precision if possible
	const bool isi = !isDbl(str, d, &ptr);

	if (isi)
		isInt(str, i, &ptr);

	if (!*ptr && isi && i == 0)
		t = Time(0,0);
	else
	if (!strcmp(ptr, "msec") || !strcmp(ptr, "ms"))
		t = isi ? Time::Msec(i) : Time::Secd(d/1e3);
	else
	if (!strcmp(ptr, "sec") || !strcmp(ptr, "s"))
		t = isi ? Time::Sec(i) : Time::Secd(d);
	else
	if (!strcmp(ptr, "min"))
		t = isi ? Time::Sec(i*60) : Time::Secd(d*60);
	else
	if (!strcmp(ptr, "hour") || !strcmp(ptr, "hr"))
		t = isi ? Time::Sec(i*60*60) : Time::Secd(d*60*60);
	else
		return false;

	return true;
}

bool isSize(const char *str, Size &s) {
	// note: things are set up so that "KB" stands for "1KB" 
	//       (convenient for rates)
	const char *ptr = str;
	int i = 1;

	const bool isi = isInt(str, i, &ptr);

	if (isi && !*ptr && i == 0)
		s = 0;
	else
	if (!strcmp(ptr, "byte") || !strcmp(ptr, "bytes") || !strcmp(ptr, "B"))
		s = i;
	else
	if (!strcmp(ptr, "KB") || !strcmp(ptr, "K"))
		s = Size::KB(i);
	else
	if (!strcmp(ptr, "MB"))
		s = Size::MB(i);
	else
		return false;

	return true;
}

/* Bool */

BoolOpt::BoolOpt(OptGrp *aGrp, const char *aName, const char *aDescr, bool def):
	Opt(aGrp, aName, aDescr), theVal(def) {
}

bool BoolOpt::parse(const char *name, const char *val) {
	if (val) {
		if (!strcmp(val, "1") || !strcmp(val, "yes") || !strcmp(val, "y") || !strcmp(val, "on")) {
			theVal = true;
			return true;
		} else
		if (!strcmp(val, "0") || !strcmp(val, "no") || !strcmp(val, "n") || !strcmp(val, "off")) {
			theVal = false;
			return true;
		}
	}
	cerr << ThePrgName << ": boolean (1|0|yes|no|on|off) value expected for the `" << name << "` option; got `" << valStr(val) << "'" << endl;
	return false;
}

void BoolOpt::report() const {
	Comment << (theName[0] == 'd' ? (theVal ? "yes" : "no") : (theVal ? "on" : "off"));
}

/* Int */

IntOpt::IntOpt(OptGrp *aGrp, const char *aName, const char *aDescr, int def):
	Opt(aGrp, aName, aDescr), theVal(def) {
}

bool IntOpt::parse(const char *name, const char *val) {
	if (val) {
		if (!strcmp(val, "none")) {
			theVal = -1;
			return true;
		} else
		if (!strcmp(val, "max")) {
			theVal = 0x7fffffff;
			return true;
		} else {
			int h;
			const char *ptr = 0;
			if (isInt(val, h, &ptr) && !*ptr) { // no garbage at the end
				theVal = h;
				return true;
			}
		}
	}

	cerr << ThePrgName << ": integer value expected for the `" << name << "` option; got `" << valStr(val) << "'" << endl;
	return false;
}

void IntOpt::report() const {
	Comment << theVal;
}


/* Double */

DblOpt::DblOpt(OptGrp *aGrp, const char *aName, const char *aDescr, double def):
	Opt(aGrp, aName, aDescr), theVal(def) {
}

bool DblOpt::parse(const char *name, const char *val) {
	if (val) {
		if (!strcmp(val, "none")) {
			theVal = -1;
			return true;
		} else {
			double h;
			const char *ptr = 0;
			if (isNum(val, h, &ptr)) {
				if (!*ptr) { // nothing at the end
					theVal = h;
					return true;
				} else
				if (!strcmp(ptr, "%") || !strcmp(ptr, "p")) { // percent
					theVal = h / 100;
					return true;
				}
			}
		}
	}
	
	cerr << ThePrgName << ": double/ratio value (e.g., `0.23', `23%', or `23p') expected for the `" << name << "` option; got `" << valStr(val) << "'" << endl;
	return false;
}

void DblOpt::report() const {
	Comment << theVal;
}


/* Time */

TimeOpt::TimeOpt(OptGrp *aGrp, const char *aName, const char *aDescr, Time def):
	Opt(aGrp, aName, aDescr), Time(def), theVal(*this) {
}

bool TimeOpt::parse(const char *name, const char *val) {
	if (val) {
		if (!strcmp(val, "none")) {
			theVal = Time();
			return true;
		} else
		if (!strcmp(val, "inf")) {
			theVal = Time(0x7fffffff, 0);
			return true;
		} else
		if (isTime(val, theVal))
			return true;
	}
	cerr << ThePrgName << ": time value (e.g., '30msec') expected for the `" << name << "` option; got `" << valStr(val) << "'" << endl;
	return false;

}

void TimeOpt::report() const {
	Comment << theVal;
}


/* Size */

SizeOpt::SizeOpt(OptGrp *aGrp, const char *aName, const char *aDescr, Size def):
	Opt(aGrp, aName, aDescr), Size(def), theVal(*this) {
}

bool SizeOpt::parse(const char *name, const char *val) {
	if (val) {
		if (!strcmp(val, "none")) {
			theVal = Size();
			return true;
		} else
		if (!strcmp(val, "inf")) {
			theVal = 0x7fffffff;
			return true;
		} else
		if (isSize(val, theVal))
			return true;
	}
	cerr << ThePrgName << ": size value (e.g., '10KB') expected for the `" << name << "` option; got `" << valStr(val) << "'" << endl;
	return false;

}

void SizeOpt::report() const {
	Comment << theVal;
}


/* String */

StrOpt::StrOpt(OptGrp *aGrp, const char *aName, const char *aDescr, const char *def = 0):
	Opt(aGrp, aName, aDescr), String(def), theVal(*this) {
}

bool StrOpt::parse(const char *name, const char *val) {
	if (val) {
		if (!strcmp(val, "unique"))
			theVal = Id2Str(UniqId::Create());
		else
			theVal = val;
		return true;
	}
	cerr << ThePrgName << ": string value expected for the `" << name << "` option; got `" << valStr(val) << "'" << endl;
	return false;
}

void StrOpt::report() const {
	Comment << valStr(theVal);
}


/* NetAddr */

NetAddrOpt::NetAddrOpt(OptGrp *aGrp, const char *aName, const char *aDescr, const NetAddr &def):
	Opt(aGrp, aName, aDescr), theVal(def) {
}

bool NetAddrOpt::parse(const char *name, const char *val) {
	String pref, suf;
	if (val && splitVal(val, pref, suf)) {
		int p;
		if (isInt(suf, p) && p >= 0) {
			theVal.addr(pref);
			theVal.port(p);
			return true;
		} else
			cerr << ThePrgName << ": invalid port specification for the `" << name << "` option; got `" << valStr(val) << "'" << endl;
	} else
		cerr << ThePrgName << ": net address (<host>:<port>) value expected for the `" << name << "` option; got `" << valStr(val) << "'" << endl;
	return false;
}

void NetAddrOpt::report() const {
	Comment << theVal;
}


/* Rate */

RateOpt::RateOpt(OptGrp *aGrp, const char *aName, const char *aDescr):
	Opt(aGrp, aName, aDescr), theCount(0) {
}

bool RateOpt::parse(const char *name, const char *val) {
	String pref, suf;
	if (val && splitVal(val, pref, suf, '/')) {
		if (isNum(pref, theCount) && theCount >= 0 && isTime(suf, theInterval))
			return true;
	}
	cerr << ThePrgName << ": rate value (e.g. `5/sec') expected for the `" << name << "` option; got `" << valStr(val) << "'" << endl;
	return false;
}

double RateOpt::rate() const {
	return theInterval > 0 ? Ratio(theCount, theInterval.secd()) : -1;
}

Time RateOpt::intArrTime() const {
	return theCount > 0 ? theInterval/theCount : Time();
}

void RateOpt::report() const {
	Comment << rate() << "/sec";
}


/* Distr */

DistrOpt::DistrOpt(OptGrp *aGrp, const char *aName, const char *aDescr, RndDistr *def):
	Opt(aGrp, aName, aDescr), RndDistr(new RndGen), theVal(def) {
}

// format: type(p1,p2,...) or type:p1,p2,...
bool DistrOpt::parse(const char *name, const char *val) {
	if (val) {
		delete theVal; // destroy default if any
		theVal = 0;

		if (!strcmp(val, "none"))
			return true;

		// now must set custom distribution

		String v = strchr(val, ')') ? String(val)(0,strlen(val)-1) : (String)val;
		String type, params;

		if (splitVal(v, type, params, ':') || splitVal(v, type, params, '(')) {
			String s[2];
			double p[2];

			// seed can be changed later
			if (type == "unif") {
				if (splitVal(params, s[0], s[1], ',') &&
					isRT(s[0], p[0]) && isRT(s[1], p[1]))
					theVal = new UnifDistr(theGen, p[0], p[1]);
			} else
			if (type == "exp") {
				if (isRT(params, p[0]))
					theVal = new ExpDistr(theGen, p[0]);
			} else
			if (type == "norm") {
				if (splitVal(params, s[0], s[1], ',') &&
					isRT(s[0], p[0]) && isRT(s[1], p[1]))
					theVal = new NormDistr(theGen, p[0], p[1]);
			} else
			if (type == "const") {
				if (isRT(params, p[0]))
					theVal = new ConstDistr(theGen, p[0]);
			} else {
				cerr << ThePrgName << ": unknown distribution `" << type << "' for the `" << name << "` option; got `" << valStr(val) << "'" << endl;
				return false;
			}

			if (!theVal) {
				cerr << ThePrgName << ": bad parameters for `" << type << "' distribution in the `" << name << "` option; got `" << valStr(val) << "'" << endl;
				return false;
			}

			return true;
		}
	}
	cerr << ThePrgName << ": distribution value (e.g. `norm(3,1)' or `norm:3,1') expected for the `" << name << "` option; got `" << valStr(val) << "'" << endl;
	return false;
}

bool DistrOpt::isRT(const String &param, double &val) const {
	if (theArgType == "num")
		return isNum(param, val);

	if (theArgType == "time") {
		Time t;
		if (isTime(param, t)) {
			val = t.secd();
			return true;
		}
		return false;
	}
	if (theArgType == "size") {
		Size s;
		if (isSize(param, s)) {
			val = s;
			return true;
		}
		return false;
	}
	Assert(0); // unknown argument type	
	return false;
}

void DistrOpt::report() const {
	if (theVal)
		Comment << *theVal;
	else
		Comment << "<none>";
}


/* Goal */

GoalOpt::GoalOpt(OptGrp *aGrp, const char *aName, const char *aDescr):
	Opt(aGrp, aName, aDescr), theVal(*this) {
}

bool GoalOpt::parse(const char *name, const char *val) {
	if (val) {
		String s[3], tail;
		int count = 1;

		if (splitVal(val, s[0], tail)) {
			count = 2;
			if (splitVal(tail, s[1], s[2]))
				count = 3;
			else
				s[1] = tail;
		} else 
			s[0] = val;

		int dcount = 0;
		int icount = 0;
		bool valid = true;
		// first integer if any is compl count
		// second integer if any is err count
		// time if any is duration
		for (int i = 0; valid && i < count; ++i) {
			const char *ptr;
			if (isTime(s[i], theDuration))
				valid = !dcount++;
			else
			if (!icount++)
				valid = isInt(s[i], theXactComplCnt, &ptr) && !*ptr;
			else
			if (icount == 2)
				valid = isInt(s[i], theErrCnt, &ptr) && !*ptr;
			else
				valid = false;
		}
		if (valid)
			return true;
	}
	cerr << ThePrgName << ": goal value (e.g. `1000:1hr:10') expected for the `" << name << "` option; got `" << valStr(val) << "'" << endl;
	return false;
}

void GoalOpt::report() const {
	Comment << theXactComplCnt << ':' << theDuration << ':' << theErrCnt;
}


/* HelpOpt */

HelpOpt::HelpOpt(OptGrp *aGrp, const char *aName, const char *aDescr):
	Opt(aGrp, aName, aDescr) {
}

bool HelpOpt::parse(const char *, const char *) {
	Assert(theCmdLine);
	theCmdLine->usage(cerr);
	exit(0);
}


/* TwoInt */

TwoIntOpt::TwoIntOpt(OptGrp *aGrp, const char *aName, const char *aDescr):
	Opt(aGrp, aName, aDescr), theLoVal(-1), theHiVal(-2) {
}

bool TwoIntOpt::parse(const char *name, const char *val) {
	if (val) {
		String pref, suf;
		// XXX: "eph..." does not really belong to this level
		if (!strcmp(val, "none") || !strcmp(val, "ephemeral") || !strcmp(val, "eph")) {
			theLoVal = -1;
			theHiVal = -2;
			return true;
		} else
		if (splitVal(val, pref, suf)) {
			if (isInt(pref, theLoVal) && isInt(suf, theHiVal)) {
				if (theLoVal >= 0 && theHiVal >= 0 && theLoVal <= theHiVal)
					return true;
			}
		} 
	}
	cerr << ThePrgName << ": <low>:<hi> expected for the `" << name << "` option; got `" << valStr(val) << "'" << endl;
	return false;
}

void TwoIntOpt::report() const {
	if (set())
		Comment << theLoVal << ':' << theHiVal;
	else
		Comment << "<none>";
}
