
/*
 * $Id: polyclt.cc,v 1.1.1.6 1999/02/12 22:27:44 rousskov Exp $
 *
 *
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "polygraph.h"

#include <new.h>
#include <signal.h>

#include <iostream.h>
#include <fstream.h>
#include <strstream.h>
#include <iomanip.h>

#include "Clock.h"
#include "SyncClt.h"
#include "AsyncClt.h"
#include "Rnd.h"
#include "StatPhaseMgr.h"
#include "EphPortMgr.h"
#include "ExpPortMgr.h"
#include "OLog.h"
#include "LogComment.h"
#include "SharedOpts.h"
#include "Socket.h"
#include "CltOpts.h"
#include "CltConnMgr.h"
#include "World.h"

#include "gadgets.h"
#include "globals.h"
#include "logTags.h"

static bool doShutdown = false;

// runs one select loop cycle
static
void step() {
	Time timeout = TheAlarmClock.timeLeft();
	//cerr << here << " select tout: " << timeout << " AlarmClockClear: " << TheAlarmClock.clear() << endl;

	FD_Set &rdSet = TheSelect->set(dirRead);
	FD_Set &wrSet = TheSelect->set(dirWrite);

	const int readyCount = TheSelect->select(TheAlarmClock.clear() ? 0 : &timeout);
	CallStat(ThePinger, selected());
	//cerr << here << " select ready: " << readyCount << endl;

	if (readyCount > 0) {
		for (int fd = TheSelect->maxFD(); fd >= 0; --fd) {
			if (rdSet.isReady(fd))
				rdSet[fd]->noteReadReady(fd);
			if (wrSet.isReady(fd))
				wrSet[fd]->noteWriteReady(fd);
		}
	} else
	if (readyCount < 0)
		Must(errno == EINTR);

	//cerr << here << " select done, was ready: " << readyCount << endl;
}

static
void shutDown(int sig) {
	Comment << "got shut down signal (" << sig << ")" << endc;
	doShutdown = true;
}

static
void configureLogs(int prec) {
	if (TheOpts.theConsFileName && strcmp(TheOpts.theConsFileName, "-"))
		cout = cerr = clog = *new ofstream(TheOpts.theConsFileName);

	if (TheOpts.theLogFileName)
		TheLog().stream(TheOpts.theLogFileName, new ofstream(TheOpts.theLogFileName));
	TheLog().resize(TheOpts.theLogSize);
	Comment.TheEchoLevel = TheOpts.theVerbLevel;

	configureStream(cout, prec);
	configureStream(cerr, prec);
	configureStream(clog, prec);
}

// set some general stuff and
// propogate cmd line options to corresponding objects
static
void configure() {

	configureLogs(2);

	// ignore some signals
	signal(SIGPIPE, SIG_IGN);
	signal(SIGHUP, SIG_IGN);

	// quit nicely on some signals
	signal(SIGINT, &shutDown);

	// set resource limits
	Check(SetMaxFD(Select::TheFDLimit) > 0);

	Socket::IsEnabled = TheCltOpts.doIOs;

	// set random seed for DistrOpts
	// note that distributions seeded with oids do not really require this
	TheOpts.passSeed(TheOpts.theRngSeed);
	TheCltOpts.passSeed(TheOpts.theRngSeed);

	TheStatPhaseMgr.goal(TheOpts.theGoal);

	// configure World
	TheWorld.type(TheCltOpts.theWorldType);
	TheWorld.id(TheCltOpts.theWorldId);
	TheWorld.capacity(TheCltOpts.theWorldCap);
	if (TheCltOpts.theWorldUrlsFileName && !TheWorld.loadUrls(TheCltOpts.theWorldUrlsFileName)) {
		Comment << "failed to load urls from `" << TheCltOpts.theWorldUrlsFileName << "'" << endc;
		exit(-2);
	}
	if (TheCltOpts.theReqOrder >= 0) {
		Assert(TheOpts.theGoal.theXactComplCnt >= 0);
		TheWorld.genObjIds(TheOpts.theGoal.theXactComplCnt, TheCltOpts.theReqOrder, TheCltOpts.theUniqStart >= 0);
	}

	// explicit or ephemeral binding
	ThePortMgr = TheCltOpts.thePorts.set() ?
		new ExpPortMgr(TheCltOpts.thePorts.lo(), TheCltOpts.thePorts.hi()) :
		new EphPortMgr();
}

int main(int argc, char *argv[]) {

	CmdLine cmd(Array<OptGrp*>() << &TheOpts << &TheCltOpts);
	if (!cmd.parse(argc, argv) || !TheOpts.validate() || !TheCltOpts.validate())
		return -1;

	Clock::Update(false);

	configure();

	cmd.report();

	const UniqId runId = UniqId::Create();
	Comment(1) << "run-id: " << runId << " pid: " << getpid() << endc;
	TheLog() << bege(lgRunId) << runId << ende;

	CltConnMgr TheConnMgr;
	TheConnMgr.configure(TheCltOpts.thePxyAddr, 
		TheCltOpts.theConnPoolCap,
		Max(TheCltOpts.theRobotCnt * (TheCltOpts.theReqRate ? 1024 : 1),
			(int)TheCltOpts.theConnPoolCap));

	Array<Client*> TheClients((int)TheCltOpts.theRobotCnt);

	RndDistr *thinkDistr = TheOpts.theXactThinkTime.distr();
	// request rate requires its own distribution
	if (TheCltOpts.theReqRate) {
		Assert(!thinkDistr); // cannot have both (checked in validate())
		// all clients should share one r.n.g.
		RndGen *rng = new RndGen(TheOpts.theRngSeed);
		const double mean = TheCltOpts.theReqRate.intArrTime().secd();
		// all clients should share one distribution
		// do we need distribution types other than exponential here?
		thinkDistr = new ExpDistr(rng, mean);
	}

	if (TheCltOpts.theReqRate) {
		for (int c = 0; c < TheCltOpts.theRobotCnt; ++c) {
			AsyncClt *clt = new AsyncClt();
			clt->configure(&TheConnMgr, thinkDistr);
			TheClients.append(clt);
		}
	} else {
		for (int c = 0; c < TheCltOpts.theRobotCnt; ++c) {
			SyncClt *clt = new SyncClt();
			clt->configure(&TheConnMgr, thinkDistr);
			TheClients.append(clt);
		}
	}
	Assert(TheClients.count() == TheCltOpts.theRobotCnt);

	Clock::Update();
	Comment << "waiting for replies; goal: " << TheOpts.theGoal << endc;

	ThePinger.start();
	TheSoOftenPinger.start();

	TheConnMgr.start();

	for (int c = 0; c < TheClients.count(); ++c) {
		TheClients[c]->start();
	}

	while (!doShutdown && TheProgress < TheOpts.theGoal) {
		Clock::Update();
		ThePinger.checkpoint();
		TheSoOftenPinger.checkpoint();
		step();
		TheConnMgr.checkCcLevel();
	}

	Clock::Update();
	TheStatPhaseMgr.forceFlush();

	Comment << "got " << TheProgress.theXactComplCnt << " replies; roughly " 
		<< Ratio(TheProgress.theXactComplCnt, TheProgress.theDuration.secd()) << " rep/sec"
		<< endc;
	return 0;
}
