#ifndef X_RND_H
#define X_RND_H

/*
 * $Id: Rnd.h,v 1.1.1.2 1999/02/12 22:27:51 rousskov Exp $
 *
 *
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include <math.h>

class ostream;

// simple random generator
// every generator gets its own seed and state
// no virtual stuff as we do not need more than one type of generators
class RndGen {
	public:
		typedef long State;

	public:
		RndGen(State aSeed = 1) { seed(aSeed); }

		// uniform[0,1)
		double trial() { return ltrial()/2147483648.0; }
		double operator ()() { return trial(); }

		// uniform[0,2147483648)
		long ltrial();

		// equiprob[lo,hi); note that "hi" is excluded
		double operator ()(double lo, double hi) { return lo + (hi-lo)*trial(); }
		int operator ()(int lo, int hi) { return (int)(lo + (hi-lo)*trial()); }

		// change seed
		void seed(State seed);
		void reset() { seed(theSeed); }

	protected:
		State theSeed;  // initial seed
		State theState; // current state
};

// abstract distribution
// allows for sharing of one random generator among several distributions
class RndDistr
{
	public:
		RndDistr(RndGen *aGen): theGen(aGen) {}
		virtual ~RndDistr() {} // may be leaking; we do not delete theGen

		virtual const char *pdfName() const = 0; // type of pdf (e.g., exp)
		RndGen *rndGen() { return theGen; }

		// returns next pseudo random number
		double operator ()() { return trial(); }
		virtual double trial() { return theGen->trial(); }

		virtual double mean() const = 0; // mean
		virtual double sdev() const = 0;  // standard deviation
		double var() const { return sdev()*sdev(); }  // variance

		virtual ostream &print(ostream &os) const;

	protected:
		RndGen *theGen;
};

inline
ostream &operator <<(ostream &os, const RndDistr &d) { return d.print(os); }

// constant(mean) distribution
class ConstDistr: public RndDistr
{
	public:
		ConstDistr(RndGen *aGen, double aMean): RndDistr(aGen), theMean(aMean) {}

		virtual const char *pdfName() const { return "const"; }

		virtual double mean() const { return theMean; }
		virtual double sdev() const { return 0; }

		virtual double trial() { return theMean; }

	protected:
		double theMean;
};


// uniform[lo,hi) distribution
class UnifDistr: public RndDistr
{
	public:
		UnifDistr(RndGen *aGen, double aLo, double aHi);

		virtual const char *pdfName() const { return "unif"; }

		virtual double mean() const { return (theLo+theHi)/2; }
		virtual double sdev() const { return range()/sqrt(12); }
		double range() const { return theHi-theLo; }

		virtual double trial() { return theLo + range()*RndDistr::trial(); }

		virtual ostream &print(ostream &os) const;

	protected:
		double theLo;
		double theHi;
};

// exponential(mean) distribution
class ExpDistr: public RndDistr
{
	public:
		ExpDistr(RndGen *aGen, double aMean): RndDistr(aGen), theMean(aMean) {}

		virtual const char *pdfName() const { return "exp"; }

		virtual double mean() const { return theMean; }
		virtual double sdev() const { return theMean; }

		virtual double trial();

	protected:
		double theMean;
};

// normal(mean, sdev) distribution
class NormDistr: public RndDistr
{
	public:
		NormDistr(RndGen *aGen, double aMean, double aSDev): RndDistr(aGen), theMean(aMean), theSDev(aSDev) {}

		virtual const char *pdfName() const { return "norm"; }

		virtual double mean() const { return theMean; }
		virtual double sdev() const { return theSDev; }

		virtual double trial();

		virtual ostream &print(ostream &os) const;

	protected:
		double theMean;
		double theSDev;
};

// lognormal(mu, sigmaSq) distribution
class LognDistr: public NormDistr
{
	public:
		// creates using mean/sdev
		static LognDistr *ViaMean(RndGen *aGen, double aMean, double aSDev);
		// creates using mu/sigma
		static LognDistr *ViaMu(RndGen *aGen, double aMu, double aSigmaSq) { return new LognDistr(aGen, aMu, aSigmaSq); }

	public:
		virtual const char *pdfName() const { return "logn"; }

		virtual double mean() const;
		virtual double sdev() const;

		virtual double trial();

	protected:
		LognDistr(RndGen *generator, double aMu, double aSigmaSq): NormDistr(generator, aMu, aSigmaSq) {}
};

#endif
